from collections import namedtuple
from contextlib import contextmanager
from hashlib import md5
import os

import netCDF4
import numpy as np
import tiledb


class NCDataModel(object):

    def __init__(self, netcdf_filename):
        self.netcdf_filename = netcdf_filename
        self._ncds = None

        self.data_var_names = []
        self.dim_coord_names = []
        self.scalar_coord_names = []
        self.aux_coord_names = []
        self.bounds = []
        self.grid_mapping = []
        self.cell_methods = []
        self.cell_measures = []
        self.unlimited_dim_coords = []

        self.domains = []
        self.domain_varname_mapping = None
        self.varname_domain_mapping = None
        self.shape = None
        self.chunks = None

        self._nc_loaded = False
        self._classified = False

    def open(self):
        # Open the NC file and retrieve key elements of it.
        self._ncds = netCDF4.Dataset(self.netcdf_filename, mode='r')
        self.dimensions = self._ncds.dimensions
        self.variables = self._ncds.variables
        # Also set derived attributes, if not already done.
        if not self._nc_loaded:
            self.dimension_names = list(self.dimensions.keys())
            self.variable_names = list(self.variables.keys())
            self.ncattrs = {key: self._ncds.getncattr(key) for key in self._ncds.ncattrs()}
            self._nc_loaded = True

    def close(self):
        self._ncds.close()

    @contextmanager
    def open_netcdf(self):
        try:
            self.open()
            yield
        finally:
            self.close()

    def dataset_open(self):
        """Check if the dataset has been loaded and is still open."""
        result = False
        if self._ncds is not None:
            result = self._ncds.isopen()
        return result

    @property
    def data_vars_mapping(self):
        return None

    def populate(self):
        with self.open_netcdf():
            self.classify_variables()
            self.get_metadata()

    def classify_variables(self):
        """
        Classify all of the NetCDF variables as one of the following:
          * Data Variable
          * Dimension Coordinates
          * Auxiliary Coordinates
          * Scalar Coordinates
          * Grid Mapping Variable
          * Bounds
          * Cell Measures
          * (Cell Methods)
          * Unlimited Dim Coords
          * Something else.

        """
        # Classify unlimited dimensions.
        self.unlimited_dim_coords = [name for name in self.dimension_names
                                     if self.dimensions[name].isunlimited()]

        classified_vars = []
        for variable_name, variable in self.variables.items():
            # Check if this variable is a grid mapping variable.
            if hasattr(variable, 'grid_mapping_name'):
                self.grid_mapping.append(variable_name)
                classified_vars.append(variable_name)

            # Check if this variable is a data variable.
            elif hasattr(variable, 'coordinates'):
                self.data_var_names.append(variable_name)
                classified_vars.append(variable_name)
                # If it's a data variable it might also have cell methods.
                if hasattr(variable, 'cell_methods'):
                    self.cell_methods.append(variable.cell_methods)

            # Check if this variable is a coordinate - dimension or aux.
            elif hasattr(variable, 'dimensions'):
                if variable_name in self.dimension_names:
                    # This is a dimension coordinate.
                    self.dim_coord_names.append(variable_name)
                elif not len(variable.dimensions):
                    # This is a scalar coordinate.
                    self.scalar_coord_names.append(variable_name)
                elif variable_name.endswith('bnds'):
                    # Check if it's a coordinate bounds.
                    self.bounds.append(variable_name)
                else:
                    # This is an auxiliary coordinate.
                    self.aux_coord_names.append(variable_name)
                classified_vars.append(variable_name)

            # Check if it's a cell measure variable.
            elif hasattr(variable, 'cell_measures'):
                self.cell_measures.append(variable_name)
                classified_vars.append(variable_name)

        # What have we still missed?
        unclassified_vars = list(set(self.variable_names) - set(classified_vars))
        if len(unclassified_vars):
            # We're not trying again, so just print them.
            print(f'Unclassified vars: {unclassified_vars}')

        # We've now classified this NC file.
        self._classified = True

    def get_chunks(self, data_var_name, max_contiguous_dims=3):
        """
        Get chunks for a named data variable `data_var_name`.

        Chunking can be tricky as 'contiguous' is a valid NetCDF
        chunking strategy (i.e. there's only one chunk and the data is
        contiguous on disk). In this case we want the chunking to match
        the shape, which is an equivalent statement.
        One heuristic we apply is that for ndim > 3 the chunking of all
        leading dimensions is [1,] to avoid very large chunks.

        """
        data_var = self.variables[data_var_name]
        chunks = data_var.chunking()
        if chunks == 'contiguous':
            shape = data_var.shape
            data_ndim = len(shape)
            overflow_dims = data_ndim - max_contiguous_dims
            if data_ndim > max_contiguous_dims:
                # More than 3D so chunk along outer (leading) dimension
                # to keep chunk sizes down.
                leading_chunksizes = [1] * overflow_dims
                chunks = tuple(leading_chunksizes + list(shape[overflow_dims:]))
            else:
                chunks = shape
        return chunks

    def get_domains(self):
        """
        Determine the unique set of domains described in all the variables in the dataset.
        Here domain means the set of dimensions that describe a variable, such as
        (time, lon, lat), which itself is a super-domain of the domain (lon, lat). As such
        any variable on the domain (lon, lat) can be fitted into the super-domain.

        If there is only one variable, this can be skipped.

        """
        # Work out the shapes of each variable. The most enclosing domains will have the highest ndim.
        ndims = np.array([len(self.variables[var_name].shape) for var_name in self.data_var_names])
        max_ndim = max(ndims)
        # Get the variables that describe the most enclosing domains (super domains).
        super_domain_vars = np.array(self.data_var_names)[ndims == max_ndim]
        domain_dims = [self.variables[var_name].dimensions for var_name in super_domain_vars]
        # Get the unique super domains.
        super_domains = list(set(domain_dims))
        # Get the variables that haven't been checked for domain inclusion.
        undomained_vars = set(self.data_var_names) - set(super_domains)

        # Check for super domains with fewer than the maximum ndim.
        for var_name in self.data_var_names:
            dims = self.variables[var_name].dimensions
            partial_coverage = [set(dims) - set(domain) for domain in super_domains]
            if all(partial_coverage):
                # This particular domain isn't fully represented by any existing super-domain,
                # so is its own new super-domain.
                super_domains.append(dims)

        self.domains = super_domains

    def domain_for_var(self):
        """
        Produce a mapping between each variable name and the super-domain that describes it.

        TODO: the first super-domain will be chosen. For a 2D domain this could result in
        incorrect classification if the first matching super-domain doesn't match the scalar
        coords of the variable.

        """
        # Build a dictionary mapping var_name to covered dimensions for undomained variables.
        name_dims_mapping = {var_name: self.variables[var_name].dimensions
                             for var_name in self.data_var_names}

        # Filling this is our target.
        name_domain_mapping = {k: [] for k in self.domains}
        # TODO: use an itertools product instead.
        for var_name, dims in name_dims_mapping.items():
            valid_domain = None
            for domain in self.domains:
                # Check if all the dims are included in the domain. If they are, then...
                if len(set(dims) - set(set(dims) & set(domain))) == 0:
                    # ... this is a valid super domain.
                    valid_domain = domain
                    break
            name_domain_mapping[valid_domain].append(var_name)
        self.domain_varname_mapping = name_domain_mapping

    def get_metadata(self):
        """
        Set extra metadata now that we've classified the dataset variables, notably:

          * the domain(s)
          * chunks
          * shape

        The last two are only done if there is just a single variable, otherwise they are
        left as `None`, and downstream writer class is responsible for setting these on a
        per-domain or per-data_variable basis.

        """
        n_data_vars = len(self.data_var_names)
        if n_data_vars == 1:
            # Only one data var so we can set domain, chunks and shape from it.
            data_var = self.variables[self.data_var_names[0]]
            self.domains.append(data_var.dimensions)
            self.domain_varname_mapping = {data_var.dimensions: [self.data_var_names[0]]}
            self.chunks = self.get_chunks(self.data_var_names[0])
            self.shape = data_var.shape
        elif n_data_vars > 1:
            # Multiple data vars means we need to set domains and variable-domain mapping.
            self.get_domains()
            self.domain_for_var()
        else:
            # No data var means trouble.
            raise ValueError(f'Expected to find at least one data var, but found {n_data_vars}.')

        # Also create the inverse mapping of var_name --> domain.
        self.varname_domain_mapping = {vi: k for k, v in self.domain_varname_mapping.items() for vi in v}


def metadata_hash(data_model, name):
    """
    Produce a completely predictable hash from the metadata of a data model.
    This will make it possible to correctly index an existing array to append
    further data onto the correct index of the existing array.

    The predictable metadata hash of the array is of the form:
        ```name(s)_hash```

    where:
        * `name` is the name(s) of the dataset (CF standard_name if available)
        * `hash` is an md5 hash of a standardised subset of the data model's metadata.

    The metadata that makes up the hash is as follows:
        * name of data variable
        * shape of dataset
        * dimension coordinate names
        * grid_mapping name
        * string of cell methods applied to dataset.

    """
    dims = ",".join(data_model.dim_coord_names)
    grid_mapping = ",".join(data_model.grid_mapping)
    cell_methods = ",".join(data_model.cell_methods)

    to_hash = f"{name}_{dims}_{data_model.shape}_{grid_mapping}_{cell_methods}"
    metadata_hash = md5(to_hash.encode("utf-8")).hexdigest()

    return f"{name}_{metadata_hash}"


class _VarDimLookup(object):
    def __init__(self,
                 primary_data_model,
                 data_vars_mapping=None,
                 variables=True):
        self.primary_data_model = primary_data_model
        self.data_vars_mapping = data_vars_mapping
        self.variables = variables
        self.data_model_attr = "variables" if self.variables else "dimensions"

        self._data_var_names = None

    @property
    def data_var_names(self):
        if self._data_var_names is None:
            if self.data_vars_mapping is not None:
                self.data_var_names = list(self.data_vars_mapping.keys())
            else:
                self.data_var_names = []
        return self._data_var_names

    @data_var_names.setter
    def data_var_names(self, value):
        self._data_var_names = value

    def __getitem__(self, keys):
        if keys in self.data_var_names:
            target = self.data_vars_mapping[keys]
        else:
            result = self.primary_data_model
            target = [result, keys]

        target_data_model, original_key = target
        return getattr(target_data_model, self.data_model_attr)[original_key]


class NCDataModelGroup(object):
    """
    Combine multiple data model instances (each with only a single data variable)
    into an amalgam data model containing multiple data variables.

    It is assumed (but not currently programatically confirmed) that all data
    model instances are otherwise identical (that is, all other metadata matches),
    and only the data variable changes between instances. Failure to heed this
    limitation will likely lead to broken TileDB / Zarr arrays.

    """
    def __init__(self, data_models):
        self.data_models = data_models

        self.verify()

        self.primary_data_model = self.data_models[0]
        self.netcdf_filename = self.primary_data_model.netcdf_filename

        self._data_var_names = None
        self._data_vars_mapping = None

        self.variables = _VarDimLookup(self.primary_data_model,
                                       data_vars_mapping=self.data_vars_mapping)
        self.dimensions = _VarDimLookup(self.primary_data_model,
                                        variables=False)

    def __getattr__(self, name):
        """
        Pass on all unhandled attribute get requests to the primary data model,
        which is assumed to be representative of all encapsulated data models
        with regard to other metadata.

        """
        return getattr(self.primary_data_model, name)

    @property
    def data_var_names(self):
        if self._data_var_names is None:
            self.data_var_names = list(self.data_vars_mapping.keys())
        return self._data_var_names

    @data_var_names.setter
    def data_var_names(self, value):
        self._data_var_names = value

    @property
    def data_vars_mapping(self):
        if self._data_vars_mapping is None:
            self.data_vars_mapping = self._map_data_vars()
        return self._data_vars_mapping

    @data_vars_mapping.setter
    def data_vars_mapping(self, value):
        self._data_vars_mapping = value

    def _map_data_vars(self):
        """Create a mapping of data variable names to the data model supplying that data variable."""
        dv_mapping = {}
        for dm in self.data_models:
            for name in dm.data_var_names:
                hashed_name = metadata_hash(dm, name)
                dv_mapping[hashed_name] = [dm, name]
        return dv_mapping

    @contextmanager
    def open_netcdf(self):
        try:
            self.open()
            yield
        finally:
            self.close()

    def open(self):
        for data_model in self.data_models:
            data_model.open()

    def close(self):
        for data_model in self.data_models:
            data_model.close()

    def dataset_open(self):
        """
        This 'dataset' (specifically a group of datasets) can only be considered
        'open' if all the datasets that comprise it are open.

        """
        return all([dm.dataset_open() for dm in self.data_models])

    def verify(self):
        """Not implemented!"""
        pass


from itertools import chain
import os
import warnings

import cf_units
import dask.array as da
from iris.coords import CellMethod, DimCoord
from iris.cube import Cube, CubeList
from iris.fileformats.netcdf import parse_cell_methods
import numpy as np
import tiledb
import xarray as xr
import zarr

from .grid_mappings import GridMapping
from .proxy import TileDBDataProxy
from . import utils


# Ref Iris: https://github.com/SciTools/iris/blob/master/lib/iris/_cube_coord_common.py#L75
IRIS_FORBIDDEN_KEYS = set([
        "standard_name",
        "long_name",
        "units",
        "bounds",
        "axis",
        "calendar",
        "leap_month",
        "leap_year",
        "month_lengths",
        "coordinates",
        "grid_mapping",
        "climatology",
        "cell_methods",
        "formula_terms",
        "compress",
        "add_offset",
        "scale_factor",
        "_FillValue",
    ])


class Reader(object):
    """
    Abstract reader class that defines the API.

    TODO replace all os usages with tiledb ls'.

    """
    horizontal_coord_names = ['latitude',
                              'longitude',
                              'grid_latitude',
                              'grid_longitude',
                              'projection_x_coordinate',
                              'projection_y_coordinate']
    def __init__(self, array_filepath):
        self.array_filepath = array_filepath

        self._artifact = None

    @property
    def artifact(self):
        return self._artifact

    @artifact.setter
    def artifact(self, value):
        self._artifact = value

    def to_iris(self):
        """Convert the input to an Iris cube or cubelist, depending on input."""
        raise NotImplementedError

    def to_xarray(self):
        """Convert the input to an Xarray dataset."""
        raise NotImplementedError


class TileDBReader(Reader):
    def __init__(self, array_name, array_filepath=None, container=None,
                 storage_options=None, data_array_name=None, ctx=None):
        super().__init__(array_filepath)

        self.array_name = array_name
        self.container = container
        self.storage_options = storage_options
        self.data_array_name = data_array_name
        self.ctx = ctx

        self.groups = {}
        self._arrays = None

        # Need either a local filepath or a remote container.
        utils.ensure_filepath_or_container(self.array_filepath, self.container)
        self.array_path = utils.filepath_generator(self.array_filepath,
                                                   self.container,
                                                   self.array_name,
                                                   ctx=self.ctx)

    @property
    def arrays(self):
        if self._arrays is None:
            self._array_paths()
        return self._arrays

    @arrays.setter
    def arrays(self, value):
        self._arrays = value

    def check_groups(self):
        if not len(self.groups.keys()):
            self.get_groups_and_arrays()

    def _get_array_attrs(self, array_path):
        with tiledb.open(array_path, "r", ctx=self.ctx) as A:
            nattr = A.schema.nattr
            attr_names = [A.schema.attr(i).name for i in range(nattr)]
        return attr_names

    def _array_paths(self):
        """Produce a mapping of array name to array path irrespective of groups."""
        self.check_groups()
        all_paths = chain.from_iterable([paths for paths in self.groups.values()])
        arrays = {}
        for path in all_paths:
            if path.endswith('/'):
                _, array_name = os.path.split(path[:-1])
            else:
                _, array_name = os.path.split(path)
            if array_name == self.data_array_name:
                attr_names = self._get_array_attrs(path)
                for attr_name in attr_names:
                    arrays[attr_name] = path
            else:
                # XXX assumes that we will not have duplicated array names.
                arrays[array_name] = path
        self.arrays = arrays

    def classifier(self, item_path, item_type):
        """
        Store and classify items in a `tiledb.walk` operation as either a TileDB group
        or a TileDB array in a given TileDB group.

        """
        if item_type == 'group':
            self.groups[item_path] = []
        else:
            if item_path.endswith('/'):
                group, _ = os.path.split(item_path[:-1])
            else:
                group, _ = os.path.split(item_path)
            self.groups[group].append(item_path)

    def get_groups_and_arrays(self):
        tiledb.walk(self.array_path.basename, self.classifier, ctx=self.ctx)

    def tdb_dir_contents(self, dir):
        contents = []
        tiledb.ls(self.array_path.basename, lambda obj_path, _: contents.append(obj_path),
                  ctx=self.ctx)
        return contents

    def _get_dim_coords(self, array_filepath):
        """Get the dimension describing coordinates from a TileDB array."""
        with tiledb.open(array_filepath, 'r', ctx=self.ctx) as A:
            dims_string = A.meta['dimensions']
        return dims_string.split(',')

    def _map_coords_inds(self, group_name, dim_name, spatial_inds):
        """
        Map coordinate point values for a named dimension to indices in that
        dimension and return as a slice of indices.

        TODO handle time coordinates (by loading the calendar from the array meta?).

        """
        dim_filepath = self.array_path.construct_path(group_name, dim_name)
        with tiledb.open(dim_filepath, 'r', ctx=self.ctx) as D:
            coord_points = D[:]

        # XXX won't handle repeated values (which should never appear in a dim coord).
        index_inds = [coord_points.index(si) for si in spatial_inds]

        n_inds = len(index_inds)
        if n_inds == 1:
            result_slice = slice(index_inds, index_inds+1)
        else:
            max_inds = 3  # [start, stop, step]
            if n_inds < max_inds:
                index_inds += [None] * (max_inds-n_inds)
            result_slice = slice(*index_inds)
        return result_slice

    def spatial_index(self, group_name, array_name, spatial_inds):
        """
        Index a specified array in coordinate space rather than in index space.
        TileDB arrays are all described in index space, with named `Dim`s
        describing a `Domain` that encapsulates the array. Earth system data, however,
        typically is described as labelled arrays, with a named coordinate describing
        each dimension of the array.

        Practically, this provides a mapping from spatial indices (the input) to
        index space, which is used to index the array.

        NOTE: only spatial coordinate *values* are supported; datetimes in particular
        are not currently supported.

        """
        array_filepath = self.array_path.construct_path(group_name, array_name)
        array_dims = self._get_dim_coords(array_filepath)

        # Check that all the coords being spatially indexed are in the array's coords.
        coord_names = list(spatial_inds.keys())
        assert list(set(coord_names) & set(array_dims)) == coord_names

        indices = []
        for dim_name in array_dims:
            coord_vals = spatial_inds.get(dim_name, None)
            if coord_vals is None:
                indices.append(slice(None))
            else:
                dim_slice = self._map_coords_inds(group_name, dim_name, coord_vals)
                indices.append(dim_slice)

        with tiledb.open(array_filepath, 'r', ctx=self.ctx) as A:
            subarray = A[tuple(indices)]
        return subarray

    def _extract(self, array_name):
        """
        Return the path to a named array, plus paths for all the associated
        dimension arrays.

        Handles multi-attr arrays by scanning all attrs in arrays that match the data
        array name passed to `self` at instantiation.

        """
        # Sanity check the requested array name is in this TileDB.
        assert array_name in self.arrays.keys()
        named_array_path = self.arrays[array_name]

        named_group_path, _ = os.path.split(named_array_path)
        named_group_arrays = self.groups[named_group_path]

        with tiledb.open(named_array_path, 'r', ctx=self.ctx) as A:
            dim_names = A.meta['dimensions'].split(',')

        dim_paths = []
        for dim_name in dim_names:
            for array_path in named_group_arrays:
                array_path = array_path[:-1] if array_path.endswith('/') else array_path
                if array_path.endswith(dim_name):
                    dim_paths.append(array_path)
                    break
        # Confirm we have an array path for each dim_name.
        assert len(dim_paths) == len(dim_names)

        return named_array_path, dim_paths

    def _array_shape(self, nonempty_domain, slices=False):
        """
        Use the TileDB array's nonempty domain (i.e. the domain that encapsulates
        all written cells) to set the shape of the data to be read out of the
        TileDB array.

        """
        # We need to include the stop index, not exclude it.
        if slices:
            slices = [slice(start, stop+1, 1) for (start, stop) in nonempty_domain]
            return tuple(slices)  # Can only index with tuple, not list.
        else:
            # TileDB describes the nonempty domain quite annoyingly!
            return [filled[1]+1 for filled in nonempty_domain]

    def _handle_attributes(self, attrs, exclude_keys=None):
        """
        Iris contains a list of attributes that may not be written to a cube/coord's
        attributes dictionary. If any of thes attributes are present in a
        TileDB array's `meta`, remove them.

        Optionally also remove extra spurious keys defined with `exclude_keys` - such as
        dictionary items set by the writer (including `dataset` and `dimensions`.)

        """
        attrs_keys = set(attrs.keys())
        if exclude_keys is not None:
            allowed_keys = list(attrs_keys - IRIS_FORBIDDEN_KEYS - set(exclude_keys))
        else:
            allowed_keys = list(attrs_keys - IRIS_FORBIDDEN_KEYS)
        return {k: attrs[k] for k in allowed_keys}

    def _from_tdb_array(self, array_path, naming_key,
                        array_name=None, to_dask=False, handle_nan=None):
        """Retrieve data and metadata from a specified TileDB array."""
        with tiledb.open(array_path, 'r', ctx=self.ctx) as A:
            metadata = {k: v for k, v in A.meta.items()}
            if array_name is None:
                array_name = metadata[naming_key]
            if to_dask:
                schema = A.schema
                dtype = schema.attr(array_name).dtype
                chunks = [schema.domain.dim(i).tile for i in range(schema.ndim)]
                array_shape = self._array_shape(A.nonempty_domain())
                proxy = TileDBDataProxy(array_shape, dtype, array_path, array_name,
                                        handle_nan=handle_nan, ctx=self.ctx)
                points = da.from_array(proxy, chunks, name=naming_key)
            else:
                array_inds = self._array_shape(A.nonempty_domain(), slices=True)
                points = A[array_inds][array_name]
        return metadata, points

    def _load_dim(self, dim_path, grid_mapping):
        """
        Create an Iris DimCoord from a TileDB array describing a dimension.

        # TODO not handled here: circular.

        """
        metadata, points = self._from_tdb_array(dim_path, 'coord')

        coord_name = metadata.pop('coord')
        standard_name = metadata.pop('standard_name', None)
        long_name = metadata.pop('long_name', None)
        var_name = metadata.pop('var_name', None)

        # Check if we've a known horizontal coord name in order to write the
        # grid mapping as it's coord system.
        if standard_name in self.horizontal_coord_names:
            coord_system = grid_mapping
        else:
            coord_system = None

        units = metadata.pop('units')
        if coord_name == 'time':
            # Handle special-case complicated time coords.
            calendar = metadata.pop('calendar')
            units = cf_units.Unit(units, calendar=calendar)

        safe_attrs = self._handle_attributes(metadata)
        coord = DimCoord(points,
                         standard_name=standard_name,
                         long_name=long_name,
                         units=units,
                         attributes=safe_attrs,
                         coord_system=coord_system)
        return coord_name, coord

    def _load_group_dims(self, group_dim_paths, grid_mapping):
        """Load all dimension-describing (coordinate) arrays in the group."""
        group_dims = {}
        for dim_path in group_dim_paths:
            name, coord = self._load_dim(dim_path, grid_mapping)
            group_dims[name] = coord
        return group_dims

    def _load_data(self, array_path, group_dims, grid_mapping,
                   attr_name=None, separator='__', handle_nan=None):
        """
        Create an Iris cube from a TileDB array describing a data variable and
        pre-loaded dimension-describing coordinates.

        TODO not handled here: aux coords and dims, cell measures, aux factories.

        """
        single_attr_name = 'dataset'
        if attr_name is None:
            attr_metadata, lazy_data = self._from_tdb_array(array_path, single_attr_name,
                                                            to_dask=True, handle_nan=handle_nan)
            metadata = attr_metadata
            attr_name = metadata.pop(single_attr_name)
        else:
            attr_metadata, lazy_data = self._from_tdb_array(array_path,
                                                            single_attr_name,
                                                            array_name=attr_name,
                                                            to_dask=True,
                                                            handle_nan=handle_nan)
            metadata = {}
            for key, value in attr_metadata.items():
                # Varname-specific keys are of form `keyname__attrname`; we only want `keyname`.
                # TODO pass the separator character to the method.
                try:
                    key_name, key_attr = key.split(separator)
                    if key_attr == attr_name:
                        metadata[key_name] = value
                except ValueError:
                    # Not all keys are varname-specific; we want all of these.
                    metadata[key] = value

        cell_methods = parse_cell_methods(metadata.pop('cell_methods', None))
        dim_names = metadata.pop('dimensions').split(',')
        # Dim Coords And Dims (mapping of coords to cube axes).
        dcad = [(group_dims[name], i) for i, name in enumerate(dim_names)]
        safe_attrs = self._handle_attributes(metadata,
                                             exclude_keys=['dataset', 'multiattr', 'grid_mapping'])
        std_name = metadata.pop('standard_name', None)
        long_name = metadata.pop('long_name', None)
        var_name = metadata.pop('var_name', None)
        if all(itm is None for itm in [std_name, long_name, var_name]):
            long_name = attr_name

        cube = Cube(lazy_data,
                    standard_name=std_name,
                    long_name=long_name,
                    var_name=var_name,
                    units=metadata.pop('units', '1'),
                    dim_coords_and_dims=dcad,
                    cell_methods=cell_methods,
                    attributes=safe_attrs)
        cube.coord_system = grid_mapping
        return cube

    def _load_group_arrays(self, data_paths, group_dims, grid_mapping, handle_nan=None):
        """Load all data-describing (cube) arrays in the group."""
        cubes = []
        for data_path in data_paths:
            cube = self._load_data(data_path, group_dims, grid_mapping, handle_nan=handle_nan)
            cubes.append(cube)
        return cubes

    def _load_multiattr_arrays(self, data_paths, group_dims, grid_mapping,
                               attr_names=None, handle_nan=None):
        """Load all data-describing (cube) attrs from a multi-attr array."""
        if isinstance(attr_names, str):
            attr_names = [attr_names]

        cubes = []
        for data_path in data_paths:
            if attr_names is None:
                with tiledb.open(data_path, 'r', ctx=self.ctx) as A:
                    attr_names = A.meta['dataset'].split(',')
            for attr_name in attr_names:
                cube = self._load_data(data_path, group_dims, grid_mapping,
                                       attr_name=attr_name, handle_nan=handle_nan)
                cubes.append(cube)
        return cubes

    def _get_grid_mapping(self, data_array_path):
        """
        Get the grid mapping (Iris coord_system) from the data array metadata.
        Grid mapping is stored as a JSON string in the array meta,
        which is translated by `.grid_mappings.GridMapping`.

        """
        grid_mapping = None
        with tiledb.open(data_array_path, 'r', ctx=self.ctx) as A:
            try:
                grid_mapping_str = A.meta['grid_mapping']
            except KeyError:
                grid_mapping_str = None
        if grid_mapping_str is not None and grid_mapping_str != 'none':
            # Cannot write NoneType into TileDB array meta, so `'none'` is a
            # stand-in that must be caught.
            translator = GridMapping(grid_mapping_str)
            try:
                grid_mapping = translator.get_grid_mapping()
            except Exception as e:
                exception_type = e.__class__.__name__
                warnings.warn(f'Re-raised as warning: {exception_type}: {e}.\nGrid mapping will be None.')
        return grid_mapping

    def _get_arrays_and_dims(self, group_array_paths):
        """
        Sort the contents of a TileDB group into data arrays and dimension arrays
        by checking the metadata of each TileDB array.

        By convention of `.writers.TDBWriter.populate_array`, a data array will
        have a metadata attribute of `'dimensions'` and a dim array will have a
        metadata attribute of `'coord'`.

        """
        dim_array_paths = []
        data_array_paths = []
        for array_path in group_array_paths:
            with tiledb.open(array_path, 'r', ctx=self.ctx) as A:
                metadata = {k: v for k, v in A.meta.items()}
            if metadata.get('dataset') is not None:
                data_array_paths.append(array_path)
            elif metadata.get('coord') is not None:
                dim_array_paths.append(array_path)
            else:
                # Can't handle this!
                raise ValueError(f'Type of array at {array_path!r} not known.')
        return dim_array_paths, data_array_paths

    def _names_to_arrays(self, names):
        """Convert one or more named datasets into groups to convert to cubes."""
        if isinstance(names, str):
            names = [names]
        iter_groups = {}
        for name in names:
            # Extract a named dataset as a single Iris cube.
            named_array_path, named_array_dims = self._extract(name)
            if named_array_path.endswith('/'):
                named_array_group_path, _ = os.path.split(named_array_path[:-1])
            else:
                named_array_group_path, _ = os.path.split(named_array_path)
            if named_array_group_path in iter_groups.keys():
                iter_groups[named_array_group_path].extend(named_array_dims + [named_array_path])
            else:
                iter_groups[named_array_group_path] = named_array_dims + [named_array_path]
        result = {}
        for k, v in iter_groups.items():
            result[k] = list(set(v))
        return result

    def to_iris(self, names=None, handle_nan=None):
        """
        Convert all arrays in a TileDB into one or more Iris cubes.

        """
        self.check_groups()

        # XXX will only return the first match if more than one cube matching `name`
        # is found.
        if names is not None:
            iter_groups = self._names_to_arrays(names)
        else:
            iter_groups = self.groups

        cubes = []
        for group_path, group_array_paths in iter_groups.items():
            dim_paths, data_paths = self._get_arrays_and_dims(group_array_paths)
            grid_mapping = self._get_grid_mapping(data_paths[0])
            group_coords = self._load_group_dims(dim_paths, grid_mapping)
            if self.data_array_name is not None:
                group_cubes = self._load_multiattr_arrays(data_paths,
                                                          group_coords,
                                                          grid_mapping,
                                                          attr_names=names,
                                                          handle_nan=handle_nan)
            else:
                group_cubes = self._load_group_arrays(data_paths, group_coords, grid_mapping,
                                                      handle_nan=handle_nan)
            cubes.extend(group_cubes)

        self.artifact = cubes[0] if len(cubes) == 1 else CubeList(cubes)
        return self.artifact

    def to_xarray(self, names=None):
        intermediate = self.to_iris(names=names)
        self.artifact = xr.from_iris(intermediate)
        return self.artifact


class ZarrReader(Reader):
    def __init__(self, array_filepath):
        super().__init__(array_filepath)

    def to_iris(self):
        intermediate = self.to_xarray()
        self.artifact = intermediate.to_iris()
        return self.artifact

    def to_xarray(self):
        self.artifact = xr.open_zarr(self.array_filepath)
        return self.artifact



"""Utility functions for `tiledb_netcdf`."""


from .paths import PosixArrayPath, AzureArrayPath


def ensure_filepath_or_container(filepath, container):
    """Ensure that either a filepath or a container has been specified, but not both."""
    # Need either a local filepath or a remote container.
    if filepath is None and container is None:
        raise ValueError("Must supply one of: array filepath, azure container.")
    if filepath is not None and container is not None:
        raise ValueError("Must supply either: array filepath; azure container, but got both.")


def filepath_generator(array_filepath, container, array_name, ctx=None):
    result = None
    if array_filepath is not None:
        result = PosixArrayPath(array_filepath, array_name)
    elif container is not None:
        result = AzureArrayPath(container, array_name, ctx=ctx)
    return result


from distributed.protocol import dask_serialize, dask_deserialize
import numpy as np
import tiledb


# Inspired by https://github.com/SciTools/iris/blob/master/lib/iris/fileformats/netcdf.py#L418.
class TileDBDataProxy(object):
    """A proxy to the data of a single TileDB array attribute."""

    __slots__ = ("shape", "dtype", "path", "var_name", "ctx", "handle_nan")

    def __init__(self, shape, dtype, path, var_name, ctx=None, handle_nan=None):
        self.shape = shape
        self.dtype = dtype
        self.path = path
        self.var_name = var_name
        self.ctx = ctx
        self.handle_nan = handle_nan

    @property
    def ndim(self):
        return len(self.shape)

    def __getitem__(self, keys):
        with tiledb.open(self.path, 'r', ctx=self.ctx) as A:
            data = A[keys][self.var_name]
            if self.handle_nan is not None:
                if self.handle_nan == 'mask':
                    data = np.ma.masked_invalid(data, np.nan)
                elif type(self.handle_nan) in [int, float]:
                    data = np.nan_to_num(data, nan=self.handle_nan, copy=False)
                else:
                    raise ValueError(f'Not a valid nan-handling approach: {self.handle_nan!r}.')
        return data

    def serialize_state(self, dummy=None):
        """
        Take the current state of `self` and make it serializable.

        Note the apparently unused kwarg `dummy`. This allows `serialize_state` to be used
        as the 'default' serialization function for msgpack. For example:

        ```
        msgpack.dumps(my_data_proxy, default=my_data_proxy.serialize_state)
        ```

        In such instances, msgpack calls `default` with the object to be dumped, which makes
        no sense in this application.

        """
        state = {}
        for attr in self.__slots__:
            value = getattr(self, attr)
            if attr == "shape":
                state[attr] = [int(i) for i in value]
            elif attr == "dtype":
                state[attr] = np.dtype(value).str
            elif attr == "ctx":
                # ctx is based on a C library that doesn't pickle...
                state[attr] = value.config().dict() if value is not None else None
            else:
                state[attr] = value
        return state

    def __getstate__(self):
        """Simplify a complex object for pickling."""
        return self.serialize_state()

    def __setstate__(self, state):
        """Restore the complex object from the simple pickled dict."""
        deserialized_state = deserialize_state(state)
        for key, value in deserialized_state.items():
            if key in self.__slots__:
                setattr(self, key, value)


def deserialize_state(s_state):
    """
    Take a serialized dictionary of state and deserialize it to set state
    on a `TileDBDataProxy` instance.

    """
    d_state = {}
    for key, s_value in s_state.items():
        if key == "dtype":
            d_value = np.dtype(s_value)
        elif key == "ctx":
            d_value = tiledb.Ctx(config=tiledb.Config(s_value)) if s_value is not None else None
        else:
            d_value = s_value
        d_state[key] = d_value
    return d_state



import json

import iris.coord_systems


def store_grid_mapping(grid_mapping_variable):
    """Store the metadata for a NetCDF grid mapping variable as a JSON string."""
    ncattrs = grid_mapping_variable.ncattrs()
    return json.dumps({k: grid_mapping_variable.getncattr(k) for k in ncattrs})


# Taken from https://github.com/SciTools/iris/blob/master/lib/iris/fileformats/_pyke_rules/fc_rules_cf.krb.
class GridMapping(object):
    """
    Convert a NetCDF grid mapping variable (expressed as a JSON string) into an
    Iris coordinate system.

    """
    # Grid mapping names.
    GRID_MAPPING_ALBERS = 'albers_conical_equal_area'
    GRID_MAPPING_AZIMUTHAL = 'azimuthal_equidistant'
    GRID_MAPPING_LAMBERT_AZIMUTHAL = 'lambert_azimuthal_equal_area'
    GRID_MAPPING_LAMBERT_CONFORMAL = 'lambert_conformal_conic'
    GRID_MAPPING_LAMBERT_CYLINDRICAL = 'lambert_cylindrical_equal_area'
    GRID_MAPPING_LAT_LON = 'latitude_longitude'
    GRID_MAPPING_MERCATOR = 'mercator'
    GRID_MAPPING_ORTHO = 'orthographic'
    GRID_MAPPING_POLAR = 'polar_stereographic'
    GRID_MAPPING_ROTATED_LAT_LON = 'rotated_latitude_longitude'
    GRID_MAPPING_STEREO = 'stereographic'
    GRID_MAPPING_TRANSVERSE = 'transverse_mercator'
    GRID_MAPPING_VERTICAL = 'vertical_perspective'

    # Earth grid attributes.
    SEMI_MAJOR_AXIS = 'semi_major_axis'
    SEMI_MINOR_AXIS = 'semi_minor_axis'
    INVERSE_FLATTENING = 'inverse_flattening'
    EARTH_RADIUS = 'earth_radius'
    LON_OF_PROJ_ORIGIN = 'longitude_of_projection_origin'
    NORTH_POLE_LAT = 'grid_north_pole_latitude'
    NORTH_POLE_LON = 'grid_north_pole_longitude'
    NORTH_POLE_GRID_LON = 'north_pole_grid_longitude'

    def __init__(self, grid_mapping_string):
        self.grid_mapping_string = grid_mapping_string
        self.grid_mapping = json.loads(self.grid_mapping_string)

    def _get_ellipsoid(self):
        """Return the ellipsoid definition."""
        major = self.grid_mapping.get(self.SEMI_MAJOR_AXIS, None)
        minor = self.grid_mapping.get(self.SEMI_MINOR_AXIS, None)
        inverse_flattening = self.grid_mapping.get(self.INVERSE_FLATTENING, None)

        if major is not None and minor is not None:
            inverse_flattening = None
        if major is None and minor is None and inverse_flattening is None:
            major = self.grid_mapping.get(self.EARTH_RADIUS, None)

        return major, minor, inverse_flattening

    def build_coordinate_system(self):
        """Create a coordinate system from the grid mapping variable."""
        major, minor, inverse_flattening = self._get_ellipsoid()
        return iris.coord_systems.GeogCS(major, minor, inverse_flattening)

    def build_rotated_coordinate_system(self):
        """Create a rotated coordinate system from the grid mapping variable."""
        major, minor, inverse_flattening = self._get_ellipsoid()
        
        north_pole_latitude = self.grid_mapping.get(self.NORTH_POLE_LAT, 90.0)
        north_pole_longitude = self.grid_mapping.get(self.NORTH_POLE_LON, 0.0)
        north_pole_grid_lon = self.grid_mapping.get(self.NORTH_POLE_GRID_LON, 0.0)
        
        ellipsoid = None
        if major is not None or minor is not None or inverse_flattening is not None:
            ellipsoid = iris.coord_systems.GeogCS(major, minor, inverse_flattening)

        return iris.coord_systems.RotatedGeogCS(north_pole_latitude, north_pole_longitude,
                                                north_pole_grid_lon, ellipsoid)

    def build_mercator_coordinate_system(self):
        """Create a Mercator coordinate system from the grid mapping variable."""
        major, minor, inverse_flattening = self._get_ellipsoid()
        longitude_of_projection_origin = self.grid_mapping.get(self.LON_OF_PROJ_ORIGIN, None)

        ellipsoid = None
        if major is not None or minor is not None or inverse_flattening is not None:
            ellipsoid = iris.coord_systems.GeogCS(major, minor, inverse_flattening)

        return iris.coord_systems.Mercator(longitude_of_projection_origin, ellipsoid=ellipsoid)

    def get_grid_mapping(self):
        """
        Determine the type of grid mapping variable we have and call the
        appropriate converter.

        Note that we do not support translation of all grid mapping variables.
        Only the following are supported:
          * latitude_longitude
          * mercator
          * rotated_latitude_longitude
        
        """
        grid_mapping_name = self.grid_mapping.get('grid_mapping_name').lower()

        if grid_mapping_name == self.GRID_MAPPING_LAT_LON:
            result = self.build_coordinate_system()
        elif grid_mapping_name == self.GRID_MAPPING_ROTATED_LAT_LON:
            result = self.build_rotated_coordinate_system()
        elif grid_mapping_name == self.GRID_MAPPING_MERCATOR:
            result = self.build_mercator_coordinate_system()
        elif grid_mapping_name == self.GRID_MAPPING_ALBERS:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_AZIMUTHAL:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_LAMBERT_AZIMUTHAL:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_LAMBERT_CONFORMAL:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_LAMBERT_CYLINDRICAL:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_ORTHO:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_POLAR:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_STEREO:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_TRANSVERSE:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        elif grid_mapping_name == self.GRID_MAPPING_VERTICAL:
            # Grid mapping type not handled yet.
            raise NotImplementedError(f'Grid mapping name {grid_mapping_name} is not currently supported.')
        else:
            # Not a valid grid mapping name.
            raise ValueError(f'{grid_mapping_name!r} is not a valid grid mapping name.')

        return result



from collections import defaultdict, namedtuple
import copy
import json
import logging
import os

import numpy as np
import tiledb
import zarr

from .data_model import NCDataModel, NCDataModelGroup
from .grid_mappings import store_grid_mapping
from .paths import PosixArrayPath, AzureArrayPath
from . import utils


append_arg_list = ['other', 'domain', 'name', 'axis', 'dim',
                   'ind_stop', 'dim_stop', 'step', 'scalar',
                   'mapping', 'verbose', 'job_number', 'n_jobs',
                   'make_data_model', 'group', 'offset', 'ctx', 'logfile']
defaults = [None] * len(append_arg_list)
AppendArgs = namedtuple('AppendArgs', append_arg_list, defaults=defaults)


class Writer(object):
    """
    An abstract base for specific writers to write the contents
    of a NetCDF file to a different format.

    """
    def __init__(self, data_model,
                 array_filepath=None, container=None, array_name=None,
                 unlimited_dims=None, ctx=None):
        self._data_model = data_model
        self.array_filepath = array_filepath
        self.container = container  # Azure container name.
        self.unlimited_dims = unlimited_dims
        self.ctx = ctx  # TileDB Context object.

        self._scalar_unlimited = None

        # Need either a local filepath or a remote container.
        utils.ensure_filepath_or_container(self.array_filepath, self.container)

        self._array_name = array_name
        if self._array_name is None:
            self.array_name = os.path.basename(os.path.splitext(self.data_model.netcdf_filename)[0])
        else:
            self.array_name = self._array_name
        self.array_path = utils.filepath_generator(self.array_filepath,
                                                   self.container,
                                                   self.array_name,
                                                   ctx=self.ctx)

    @property
    def data_model(self):
        if not self._data_model.dataset_open():
            self._data_model.open()
        return self._data_model

    def _all_coords(self, variable):
        dim_coords = list(variable.dimensions)
        other_coords = variable.coordinates.split(' ')
        return dim_coords + other_coords

    def _get_dim_coord_names(self, var_name):
        """
        Figure out names of the dimension-describing coordinates for this array,
        including the promoted append-dimension scalar coordinate if necessary.

        """
        dim_coord_names = self.data_model.variables[var_name].dimensions
        if self._scalar_unlimited is not None:
            # `dim_coord_names` is a tuple...
            dim_coord_names = (self._scalar_unlimited,) + dim_coord_names
        return dim_coord_names

    def _append_checker(self, other_data_model, var_name, append_dim):
        """Checks to see if an append operation can go ahead."""
        # Sanity checks: is the var name in both self, other, and the tiledb?
        assert var_name in self.data_model.data_var_names, \
            f'Variable name {var_name!r} not found in this data model.'
        assert var_name in other_data_model.data_var_names, \
            f'Variable name {var_name!r} not found in other data model.'

        self_var = self.data_model.variables[var_name]
        self_var_coords = self._all_coords(self_var)
        # Is the append dimension valid?
        assert append_dim in self_var_coords, \
            f'Dimension {append_dim!r} not found in this data model.'

        with other_data_model.open_netcdf():
            other_var = other_data_model.variables[var_name]
            other_var_coords = self._all_coords(other_var)
            # Is the append dimension valid?
            assert append_dim in other_var_coords, \
                f'Dimension {append_dim!r} not found in other data model.'

    def _append_dimension(self, var_name, append_desc):
        """Determine the name and index of the dimension for the append operation."""
        if not isinstance(append_desc, int):
            # Find the append axis from the dimension name.
            append_axis = self.data_model.variables[var_name].dimensions.index(append_desc)
            append_dim = append_desc
        else:
            # Find the append dimension name from the axis.
            append_axis = append_desc
            append_dim = self.data_model.dimensions[append_axis]
        return append_axis, append_dim

    def _fill_missing_points(self, coord_array_path, coord_array_name, verbose=False):
        """
        If one or more indices along the append axis are missing spatial points, we
        end up with `NaN`s in the resultant coordinate array. This prevents loading
        into Iris (as we cannot make a monotonic coordinate array).

        Fill such missing points with interpolated point values so that Iris can load
        the dataset, if with missing data points still. Use a simple custom 1D
        interpolator as the SciPy and NumPy offerings cannot handle NaN values.

        """
        with tiledb.open(coord_array_path, 'r', ctx=self.ctx) as D:
            ned = D.nonempty_domain()[0]
            coord_points = D[ned[0]:ned[1]][coord_array_name]

        missing_points, = np.nonzero(np.isnan(coord_points))
        if len(missing_points):
            if verbose:
                print(f'{len(missing_points)} points to fill in {coord_array_name!r}.')

            ind_points = np.arange(len(coord_points))
            coord_steps = np.unique(np.diff(coord_points))
            # Expects only a single non-NaN step (i.e. monotonicity).
            numeric_step, = coord_steps[np.nonzero(~np.isnan(coord_steps))]

            # Interpolate to fill the missing points.
            vec_interp = np.vectorize(fillnan)
            coord_points[missing_points] = vec_interp(ind_points[missing_points],
                                                      coord_points[0],
                                                      numeric_step)

            # Write the whole filled array back to the TileDB coord array.
            with tiledb.open(coord_array_path, 'w', ctx=self.ctx) as D:
                D[ned[0]:ned[1]] = coord_points
        else:
            if verbose:
                print(f'No missing points in {coord_array_name!r}, nothing to do.')


class _TDBWriter(Writer):
    """
    .. deprecated::
        This class is deprecated in favour of the former `MultiAttrTDBWriter`,
        now renamed to just `TDBWriter`. This preferred class provides all the
        functionality of this class but with extra functionality for writing
        multi-attr arrays as well.

        This class is being maintained for now as it provides some of the
        functionality the preferred class relies upon.

    Write Python objects loaded from NetCDF to TileDB.

    Data Model: an instance of `NCDataModel` supplying data from a NetCDF file.
    Filepath: the filepath to save the tiledb array at.

    """
    def __init__(self, data_model,
                 array_filepath=None, container=None, array_name=None,
                 unlimited_dims=None, ctx=None):
        super().__init__(data_model, array_filepath, container, array_name,
                         unlimited_dims, ctx)

        if self.unlimited_dims is None:
            self.unlimited_dims = []

    def _public_domain_name(self, domain):
        domain_index = self.data_model.domains.index(domain)
        return f'domain_{domain_index}'

    def _create_tdb_directory(self, group_dirname):
        """
        Create an on-filesystem directory for a tiledb group if it does
        not exist, and ignore the error if it does.

        """
        try:
            os.makedirs(group_dirname)
        except FileExistsError:
            pass

    def _create_tdb_dim(self, dim_name, coords):
        dim_coord = self.data_model.variables[dim_name]
        chunks = self.data_model.get_chunks(dim_name)

        # Handle scalar dimensions.
        if dim_name == self._scalar_unlimited:
            dim_coord_len = 1
            chunks = (1,)
        else:
            # TODO: work out nD coords (although a DimCoord will never be nD).
            dim_coord_len, = dim_coord.shape

        # Set the tdb dimension dtype to `int64` regardless of input.
        # Dimensions must have int indices for dense array schemas.
        # All tdb dims in a domain must have exactly the same dtype.
        dim_dtype = np.int64

        # Sort out the domain, based on whether the dim is unlimited,
        # or whether it was specified that it should be by `self.unlimited_dims`.
        if dim_name in self.unlimited_dims:
            domain_max = np.iinfo(dim_dtype).max - dim_coord_len
        elif dim_name in self.data_model.unlimited_dim_coords:
            domain_max = np.iinfo(dim_dtype).max - dim_coord_len
        else:
            domain_max = dim_coord_len

        # Modify the name of the dimension if this dimension describes the domain
        # for a dim coord array.
        # Array attrs and dimensions must have different names.
        if coords:
            dim_name = f'{dim_name}_coord'

        return tiledb.Dim(name=dim_name,
                          domain=(0, domain_max),
                          tile=chunks,
                          dtype=dim_dtype)

    def create_domain_arrays(self, domain_vars, domain_name, coords=False):
        """Create one single-attribute array per data var in this NC domain."""
        for var_name in domain_vars:
            # Set dims for the enclosing domain.

            data_var = self.data_model.variables[var_name]
            data_var_dims = data_var.dimensions
            # Handle scalar append dimension coordinates.
            if not len(data_var_dims) and var_name == self._scalar_unlimited:
                data_var_dims = [self._scalar_unlimited]
            array_dims = [self._create_tdb_dim(dim_name, coords) for dim_name in data_var_dims]
            tdb_domain = tiledb.Domain(*array_dims)

            # Get tdb attributes.
            attr = tiledb.Attr(name=var_name, dtype=data_var.dtype)

            # Create the URI for the array.
            array_filename = self.array_path.construct_path(domain_name, var_name)
            # Create an empty array.
            schema = tiledb.ArraySchema(domain=tdb_domain, sparse=False,
                                        attrs=[attr], ctx=self.ctx)
            tiledb.Array.create(array_filename, schema)

    def _array_indices(self, shape, start_index):
        """Set the array indices to write the array data into."""
        return _array_indices(shape, start_index)

    def _get_grid_mapping(self, data_var):
        """
        Get the data variable's grid mapping variable, encode it as a JSON string
        for easy storage in the TileDB array's meta and return it for storage as
        array metadata.

        """
        grid_mapping_name = data_var.getncattr("grid_mapping")
        result = 'none'  # TileDB probably won't support `NoneType` in array meta.
        if grid_mapping_name is not None:
            assert grid_mapping_name in self.data_model.grid_mapping
            grid_mapping_var = self.data_model.variables[grid_mapping_name]
            result = store_grid_mapping(grid_mapping_var)
        return result

    def populate_array(self, var_name, data_var, domain_name,
                       start_index=None, write_meta=True):
        """Write the contents of a netcdf data variable into a tiledb array."""
        # Get the data variable and the URI of the array to write to.
        var_name = data_var.name
        array_filename = self.array_path.construct_path(domain_name, var_name)

        # Write to the array.
        scalar = var_name == self._scalar_unlimited
        write_array(array_filename, data_var,
                    start_index=start_index, scalar=scalar, ctx=self.ctx)
        if write_meta:
            with tiledb.open(array_filename, 'w', ctx=self.ctx) as A:
                # Set tiledb metadata from data var ncattrs.
                for ncattr in data_var.ncattrs():
                    A.meta[ncattr] = data_var.getncattr(ncattr)
                # Add metadata describing whether this is a coord or data var.
                if var_name in self.data_model.data_var_names:
                    # A data array gets a `dataset` key in the metadata dictionary,
                    # which defines all the data variables in it.
                    A.meta['dataset'] = var_name
                    # Define this as not being a multi-attr array.
                    A.meta['multiattr'] = False
                    # Add grid mapping metadata as a JSON string.
                    grid_mapping_string = self._get_grid_mapping(data_var)
                    A.meta['grid_mapping'] = grid_mapping_string
                    # XXX: can't add list or tuple as values to metadata dictionary...
                    dim_coord_names = self._get_dim_coord_names(var_name)
                    A.meta['dimensions'] = ','.join(n for n in dim_coord_names)
                elif var_name in self.data_model.dim_coord_names:
                    # A dim coord gets a `coord` key in the metadata dictionary,
                    # value being the name of the coordinate.
                    A.meta['coord'] = self.data_model.dimensions[var_name].name
                elif var_name == self._scalar_unlimited:
                    # Handle scalar coords along the append axis.
                    A.meta['coord'] = self._scalar_unlimited
                else:
                    # Don't know how to handle this. It might be an aux or scalar
                    # coord, but we're not currently writing TDB arrays for them.
                    pass

    def populate_domain_arrays(self, domain_vars, domain_name):
        """Populate all arrays with data from netcdf data vars within a tiledb group."""
        for var_name in domain_vars:
            data_var = self.data_model.variables[var_name]
            self.populate_array(var_name, data_var, domain_name)

    def create_domains(self):
        """
        We need to create one TDB group per data variable in the data model,
        organised by domain.

        """
        for domain in self.data_model.domains:
            # Get the data and coord variables in this domain.
            domain_vars = self.data_model.domain_varname_mapping[domain]
            # Defined for the sake of clarity (each `domain` is a list of its dim coords).
            domain_coords = domain

            # Create group.
            domain_name = self._public_domain_name(domain)
            group_dirname = self.array_path.construct_path(domain_name, '')
            if self.array_filepath is not None:
                # XXX TileDB does not automatically create group directories.
                self._create_tdb_directory(group_dirname)
            tiledb.group_create(group_dirname)

            # Create and write arrays for each domain-describing coordinate.
            self.create_domain_arrays(domain_coords, group_dirname, coords=True)
            self.populate_domain_arrays(domain_coords, group_dirname)

            # Get data vars in this domain and create an array for the domain.
            self.create_domain_arrays(domain_vars, group_dirname)
            # Populate this domain's array.
            self.populate_domain_arrays(domain_vars, group_dirname)

    def append(self, others, var_name, append_dim,
              logfile=None, parallel=False, verbose=False):
        """
        Append extra data as described by the contents of `others` onto
        an existing TileDB array along the axis defined by `append_dim`.

        Notes:
          * extends one dimension only
          * cannot create new dimensions, only extend existing dimensions

        TODO support multiple axis appends.
        TODO check if there's already data at the write inds and add an overwrite?

        """
        if logfile is not None:
            logging.basicConfig(filename=logfile,
                                level=logging.DEBUG,
                                format='%(asctime)s %(message)s',
                                datefmt='%d/%m/%Y %H:%M:%S')

        make_data_model = False
        # Check what sort of thing `others` is.
        if isinstance(others, NCDataModel):
            others = [others]
        elif isinstance(others, str):
            others = [others]
            make_data_model = True
        else:
            other = others[0]
            if isinstance(other, str):
                make_data_model = True

        append_axis, append_dim = self._append_dimension(var_name, append_dim)

        self_dim_var = self.data_model.variables[append_dim]
        self_dim_points = copy.copy(self_dim_var[:])
        self_dim_start, self_dim_stop, self_step = _dim_points(self_dim_points)
        self_ind_start, self_ind_stop = _dim_inds(self_dim_points,
                                                  [self_dim_start, self_dim_stop])

        # Get domain for var_name and tiledb array path.
        domain = self.data_model.varname_domain_mapping[var_name]
        domain_name = self._public_domain_name(domain)
        domain_path = os.path.join(self.array_filepath, self.array_name, domain_name)

        # For multidim / multi-attr appends this will be more complex.
        common_job_args = [domain_path, var_name, append_axis, append_dim,
                           self_ind_stop, self_dim_stop, self_step,
                           make_data_model, verbose]
        job_args = [[other] + common_job_args for other in others]

        if parallel:
            import dask.bag as db
            bag_of_jobs = db.from_sequence(job_args)
            bag_of_jobs.map(_make_tile_helper).compute()
        else:
            for i, args in enumerate(job_args):
                args += [i, len(job_args)]
                _make_tile_helper(args)

        if len(others) > 10:
            # Consolidate at the end of the append operations to make the resultant
            # array more performant.
            config = tiledb.Config({"sm.consolidation.steps": 1000})
            ctx = tiledb.Ctx(config)
            tiledb.consolidate(os.path.join(domain_path, var_name), ctx=ctx)

    def fill_missing_points(self, append_dim_name, verbose=False):
        # XXX duplicated from `append` method.
        domain = self.data_model.varname_domain_mapping[var_name]
        domain_name = f'domain_{self.data_model.domains.index(domain)}'
        domain_path = os.path.join(self.array_filepath, self.array_name, domain_name)
        coord_array_path = os.path.join(domain_path, append_dim_name)

        self._fill_missing_points(coord_array_path, append_dim_name, verbose=verbose)


class TileDBWriter(_TDBWriter):
    """
    Write Python objects loaded from NetCDF to TileDB, including writing TileDB
    arrays with multiple data attributes from different NetCDF data variables
    that are equally dimensioned.

    Data Model: an instance of `NCDataModel` supplying data from a NetCDF file.
    Filepath: the filepath to save the tiledb array at.

    """
    def __init__(self, data_model,
                 array_filepath=None, container=None, array_name=None,
                 unlimited_dims=None, ctx=None, domain_separator=','):
        """
        Set up a writer to store the contents of one or more NetCDF data models
        in a TileDB array.

        Args:
        * data_model: An `NCDataModel` object. Defines the contents of the base NetCDF file
                      to write to TileDB.

        Kwargs:
        * array_filepath: posix-like path of location on disk to write the TileDB array.
                          Either `array_filepath` or `container` *must* be provided, but not both.
        * container: the name of an Azure Storage container to write the TileDB array to.
                     Either `array_filepath` or `container` *must* be provided, but not both.
        * array_name: the name of the root TileDB array to write. Defaults to the name of the
                      data model's NetCDF file if not set. By specifying `array_name` as a
                      relative path along with `container` you can write the TileDB array
                      to a location other than the Azure Storage container's root.
        * unlimited_dims: a named dimension that will have unlimited length in the written
                          TileDB array. Typically this is set to a dimension you wish to append to.
                          The string given here must be the name of a dimension in the supplied
                          `data_model`.
        * ctx: a TileDB Ctx (context) object, typically used to store metadata relating to the
               Azure Storage container.
        * domain_separator: a string that specifies the separator between dimensions in
                            domain names. Defaults to `,`.

        """
        super().__init__(data_model, array_filepath, container, array_name,
                         unlimited_dims, ctx)

        self.domain_separator = domain_separator
        self._domains_mapping = None

    @property
    def domains_mapping(self):
        if self._domains_mapping is None:
            self._make_shape_domains()
        return self._domains_mapping

    @domains_mapping.setter
    def domains_mapping(self, value):
        self._domains_mapping = value

    def _public_domain_name(self, dimensions, separator):
        """
        A common method for determining the domain name for a given data variable,
        based on its dimensions.

        """
        return separator.join(dimensions)

    def _get_attributes(self, data_var):
        ncattrs = data_var.ncattrs()
        metadata = {}
        for ncattr in ncattrs:
            metadata[ncattr] = data_var.getncattr(ncattr)
        # Add a fallback long_name in case of both standard_name and long_name being missing.
        if "standard_name" not in ncattrs and "long_name" not in ncattrs:
            metadata["long_name"] = data_var.name
        return metadata

    def _multi_attr_metadata(self, data_var_names):
        if len(data_var_names) == 1:
            data_var = self.data_model.variables[data_var_names[0]]
            metadata = self._get_attributes(data_var)
        else:
            metadata = {}
            for var_name in data_var_names:
                data_var = self.data_model.variables[var_name]
                data_var_meta = self._get_attributes(data_var)
                # XXX TileDB does not support dict in array metadata...
                for key, value in data_var_meta.items():
                    metadata[f'{key}__{var_name}'] = value
        return metadata

    def _make_shape_domains(self):
        """
        Make one domain for each unique combination of shape and dimension variables.

        We need to make this set of domains for the multi-attr case as a limitation
        in TileDB means that all attrs in an array must be written at the same time,
        which also means the indexing for all attrs to be written must be the same.

        """
        shape_domains = []
        for data_var_name in self.data_model.data_var_names:
            dimensions = self.data_model.variables[data_var_name].dimensions
            # Promote scalar append dimensions.
            if self.unlimited_dims not in dimensions and self.unlimited_dims in self.data_model.scalar_coord_names:
                dimensions = [self.unlimited_dims] + list(dimensions)
                self._scalar_unlimited = self.unlimited_dims
            domain_string = self._public_domain_name(dimensions, self.domain_separator)
            shape_domains.append((domain_string, data_var_name))

        domains_mapping = defaultdict(list)
        for domain, data_var_name in shape_domains:
            domains_mapping[domain].append(data_var_name)
        self.domains_mapping = domains_mapping

    def populate_multiattr_array(self, data_array_name, data_var_names, domain_name,
                                 start_index=None, write_meta=True):
        """Write the contents of multiple data variables into a multi-attr TileDB array."""
        array_filename = self.array_path.construct_path(domain_name, data_array_name)

        # Write to the array.
        data_vars = {name: self.data_model.variables[name] for name in data_var_names}
        scalar = self._scalar_unlimited is not None
        write_multiattr_array(array_filename, data_vars,
                              start_index=start_index, scalar=scalar, ctx=self.ctx)
        if write_meta:
            multi_attr_metadata = self._multi_attr_metadata(data_var_names)
            with tiledb.open(array_filename, 'w', ctx=self.ctx) as A:
                # Set tiledb metadata from data var ncattrs.
                for key, value in multi_attr_metadata.items():
                    A.meta[key] = value
                # A data array gets a `dataset` key in the metadata dictionary,
                # which defines all the data variables in it.
                A.meta['dataset'] = ','.join(data_var_names)
                # Define this as being a multi-attr array.
                A.meta['multiattr'] = True
                # Add grid mapping metadata as a JSON string.
                zeroth_data_var = list(data_vars.values())[0]
                grid_mapping_string = self._get_grid_mapping(zeroth_data_var)
                A.meta['grid_mapping'] = grid_mapping_string
                # XXX: can't add list or tuple as values to metadata dictionary...
                dim_coord_names = self._get_dim_coord_names(data_var_names[0])
                A.meta['dimensions'] = ','.join(n for n in dim_coord_names)

    def create_multiattr_array(self, domain_var_names, domain_dims,
                               domain_name, data_array_name):
        """Create one multi-attr TileDB array with an attr for each data variable."""
        # Create dimensions and domain for the multi-attr array.
        array_dims = [self._create_tdb_dim(dim_name, coords=False) for dim_name in domain_dims]
        tdb_domain = tiledb.Domain(*array_dims)

        # Set up the multiple attrs for the array.
        attrs = []
        for var_name in domain_var_names:
            dtype = self.data_model.variables[var_name].dtype
            attr = tiledb.Attr(name=var_name, dtype=dtype)
            attrs.append(attr)

        # Create the URI for the array.
        array_filename = self.array_path.construct_path(domain_name, data_array_name)
        # Create an empty array.
        schema = tiledb.ArraySchema(domain=tdb_domain, sparse=False,
                                    attrs=attrs, ctx=self.ctx)
        tiledb.Array.create(array_filename, schema)

    def create_domains(self, data_array_name='data'):
        """
        Create one TileDB domain for each unique shape / dimensions combination
        in the input Data Model. Each domain will contain:
          * one multi-attr array, where the attrs are all the data variables described
            by this combination of dimensions, and
          * one array for each of the dimension-describing coordinates for this
            combination of dimensions.

        """
        self._make_shape_domains()
        for domain_name, domain_var_names in self.domains_mapping.items():
            domain_coord_names = domain_name.split(self.domain_separator)

            # Create group.
            group_dirname = self.array_path.construct_path(domain_name, '')
            # XXX This might be failing because the TileDB root dir doesn't exist...
            # For a POSIX path we must explicitly create the group directory.
            if self.array_filepath is not None:
                # TODO why is this necessary? Shouldn't tiledb create if this dir does not exist?
                self._create_tdb_directory(group_dirname)

            tiledb.group_create(group_dirname, ctx=self.ctx)

            # Create and write arrays for each domain-describing coordinate.
            self.create_domain_arrays(domain_coord_names, domain_name, coords=True)
            self.populate_domain_arrays(domain_coord_names, domain_name)

            # Get data vars in this domain and create and populate a multi-attr array.
            self.create_multiattr_array(domain_var_names, domain_coord_names,
                                        domain_name, data_array_name)
            self.populate_multiattr_array(data_array_name, domain_var_names, domain_name)

    def _scalar_step(self, base_point, append_dim, other):
        """
        Manually calculate the append dimension point step in the scalar case
        when it cannot be done by finding the diff between successive points.

        """
        other_data_model = NCDataModel(other)
        other_data_model.classify_variables()
        offset_point = other_data_model.variables[append_dim][:]
        return offset_point - base_point

    def _get_scalar_offset(self, baseline, append_dim, self_dim_stop):
        """
        Use the specified baseline file to calcuate the single offset between every
        successive step along the scalar append dimension.

        """
        odm = NCDataModel(other)
        with odm.open_netcdf():
            odm.classify_variables()
            odm.get_metadata()
            points = np.array(odm.variables[append_dim][:])
        return points - self_dim_stop

    def _get_scalar_points_and_offsets(self, others, append_dim, self_dim_stop):
        """
        Scan all of `others` to find each offset between the existing array and
        each dataset to be appended.

        """
        odp = []
        for other in others:
            ncdm = NCDataModel(other)
            with ncdm.open_netcdf():
                ncdm.classify_variables()
                ncdm.get_metadata()
                odp.append(ncdm.variables[append_dim][:])
        other_dim_points = np.array(odp)
        offsets = other_dim_points - self_dim_stop
        return offsets.data  # Only return the non-masked element of the masked array.

    def _run_consolidate(self, domain_names, data_array_name, verbose=False):
        # Consolidate at the end of the append operations to make the resultant
        # array more performant.
        config_key_name = "sm.consolidation.steps"
        config_key_value = 100
        if self.ctx is None:
            config = tiledb.Config({config_key_name: config_key_value})
            ctx = tiledb.Ctx(config)
        else:
            cfg_dict = self.ctx.config().dict()
            cfg_dict[config_key_name] = config_key_value
            ctx = tiledb.Ctx(config=tiledb.Config(cfg_dict))
        for i, domain_name in enumerate(domain_names):
            if verbose:
                print()  # Clear last carriage-returned print statement.
                print(f'Consolidating array: {i+1}/{len(domain_names)}', end="\r")
            else:
                print('Consolidating...')
            array_path = self.array_path.construct_path(domain_name, data_array_name)
            tiledb.consolidate(array_path, ctx=ctx)

    def append(self, others, append_dim, data_array_name,
               baseline=None, group=False, logfile=None, parallel=False,
               verbose=False, consolidate=True):
        """
        Append extra data as described by the contents of `others` onto
        an existing TileDB array along the axis defined by `append_dim`.

        Notes:
          * extends one dimension only
          * cannot create new dimensions, only extend existing dimensions
          * `others` must be str of one or more files to append, not NCDataModel

        TODO support multiple axis appends.
        TODO check if there's already data at the write inds and add an overwrite?

        """
        make_data_model = False
        # Check what sort of thing `others` is.
        if isinstance(others, str):
            others = [others]

        # Check all domains for including the append dimension.
        domain_names = [d for d in self.domains_mapping.keys()
                        if append_dim in d.split(self.domain_separator)]
        domain_paths = [self.array_path.construct_path(d, '') for d in domain_names]
        append_axes = [d.split(self.domain_separator).index(append_dim) for d in domain_names]

        # Get starting dimension points and offsets.
        self_dim_var = self.data_model.variables[append_dim]
        self_dim_points = copy.copy(np.array(self_dim_var[:], ndmin=1))

        if append_dim == self._scalar_unlimited:
            if baseline is None:
                raise ValueError('Cannot determine scalar step without a baseline dataset.')
            self_ind_stop = 0
            self_dim_stop = self_dim_points[0]
            offsets = self._get_scalar_offset(baseline, append_dim, self_dim_stop)
            # offsets = self._get_scalar_points_and_offsets(others, append_dim, self_dim_stop)
            if len(offsets) == 1:
                self_step = offsets[0]
            else:
                # Smooth out any noise in slightly different offsets.
                self_step = np.median(np.diff(offsets))
            scalar = True
        else:
            self_dim_start, self_dim_stop, self_step = _dim_points(self_dim_points)
            self_ind_start, self_ind_stop = _dim_inds(self_dim_points,
                                                      [self_dim_start, self_dim_stop])
            offsets = None
            scalar = False

        # For multidim / multi-attr appends this may be more complex.
        n_jobs = len(others)
        # Prepare for serialization.
        tdb_config = self.ctx.config().dict() if self.ctx is not None else None
        all_job_args = []
        for ct, other in enumerate(others):
            offset = offsets[ct] if offsets is not None else None
            this_job_args = AppendArgs(other=other, domain=domain_paths, name=data_array_name,
                                       dim=append_dim, axis=append_axes, scalar=scalar, group=group,
                                       offset=offset, mapping=self.domains_mapping, logfile=logfile,
                                       ind_stop=self_ind_stop, dim_stop=self_dim_stop, step=self_step,
                                       verbose=verbose, job_number=ct, n_jobs=n_jobs, ctx=tdb_config)
            all_job_args.append(this_job_args)

        # Serialize to JSON for network transmission.
        serialized_jobs = map(lambda job: json.dumps(job._asdict()), all_job_args)
        if parallel:
            import dask.bag as db
            bag_of_jobs = db.from_sequence(serialized_jobs)
            bag_of_jobs.map(_make_multiattr_tile_helper).compute()
        else:
            for job_args in serialized_jobs:
                _make_multiattr_tile_helper(job_args)

        if consolidate and (n_jobs > 10):
            self._run_consolidate(domain_names, data_array_name, verbose=verbose)

    def fill_missing_points(self, append_dim_name, verbose=False):
        # Check all domains for including the append dimension.
        coord_array_paths = []
        for domain_name in self.domains_mapping.keys():
            if append_dim_name in domain_name.split(self.domain_separator):
                coord_array_path = self.array_path.construct_path(domain_name,
                                                                  append_dim_name)
                self._fill_missing_points(coord_array_path,
                                          append_dim_name,
                                          verbose=verbose)


class ZarrWriter(Writer):
    """Provides a class to write Python objects loaded from NetCDF to zarr."""
    def __init__(self, data_model, array_filepath, array_name=None):
        super().__init__(data_model, array_filepath, array_name, unlimited_dims=None)

        filename = os.path.join(os.path.abspath("."), self.array_filepath, self.array_name)
        self.array_filename = f'{filename}.zarr'
        print(self.array_filename)

        self.group = None
        self.zarray = None

    def create_variable_datasets(self, var_names):
        """
        Create a zarr group - containing data variables and dimensions - for
        a given domain.

        A domain is described by the tuple of dimensions that describe it.

        """
        # Write domain variables and dimensions into group.
        for var_name in var_names:
            data_var = self.data_model.variables[var_name]
            chunks = self.data_model.get_chunks(var_name)
            data_array = self.group.create_dataset(var_name,
                                                shape=data_var.shape,
                                                chunks=chunks,
                                                dtype=data_var.dtype)
            data_array[:] = data_var[...]

            # Set array attributes from ncattrs.
            for ncattr in data_var.ncattrs():
                data_array.attrs[ncattr] = data_var.getncattr(ncattr)

            # Set attribute to specify var's dimensions.
            data_array.attrs['_ARRAY_DIMENSIONS'] = data_var.dimensions

    def create_zarr(self):
        """
        Create a zarr for the contents of `self.data_model`. The grouped
        structure of this zarr is:

            root (filename)
             | - phenom_0
             | - phenom_1
             | - ...
             | - dimension_0
             | - dimension_1
             | - ...
             | - phenom_n
             | - ...

        TODO: add global NetCDF attributes to outermost zarr structure?

        """
        store = zarr.DirectoryStore(self.array_filename)
        self.group = zarr.group(store=store)

        # Write zarr datasets for data variables.
        for domain in self.data_model.domains:
            domain_vars = self.data_model.domain_varname_mapping[domain]
            self.create_variable_datasets(domain_vars)

        # Write zarr datasets for dimension variables.
        keys = list(self.data_model.domain_varname_mapping.keys())
        unique_flat_keys = set([k for domain in keys for k in domain])
        self.create_variable_datasets(unique_flat_keys)

    def append(self, other_data_model, var_name, append_dim):
        """
        Append the contents of other onto self.group, optionally specifying
        a single zarr array to append to with `group_name`.

        If this is not specified, extend all of the arrays in self.group_name
        with all of the arrays found in other. This assumes that the data
        variables in self and other:
          a) are identical
          b) all append along the same dimension.

        Note: append axis is limited to a single axis.

        """
        # Check if the append can go ahead.
        self._append_checker(other_data_model, var_name, append_dim)

        # Work out the index of the append dimension.
        append_axis, append_dim = self._append_dimension(var_name, append_dim)

        with other_data_model.open_netcdf():
            # Append data to the phenomenon.
            other_data_var = other_data_model.variables[var_name]
            getattr(self.group, var_name).append(other_data_var[...], axis=append_axis)

            # Append coordinate values to the append dimension.
            other_dim = other_data_model.variables[append_dim]
            getattr(self.group, append_dim).append(other_dim[...], axis=0)


###################################################################################
#                                                                                 #
# Remove these functions from `TDBWriter` because most of them are static and it  #
# might make tiling in parallel possible!                                         #
#                                                                                 #
###################################################################################


def _array_indices(shape, start_index):
    """Set the array indices to write the array data into."""
    if isinstance(start_index, int):
        start_index = [start_index] * len(shape)

    array_indices = []
    for dim_len, start_ind in zip(shape, start_index):
        array_indices.append(slice(start_ind, dim_len+start_ind))
    return tuple(array_indices)


def write_array(array_filename, data_var,
                start_index=None, scalar=False, ctx=None):
    """Write to the array."""
    if start_index is None:
        start_index = 0
        if scalar:
            shape = (1,)
        else:
            shape = data_var.shape
        write_indices = _array_indices(shape, start_index)
    else:
        write_indices = start_index

    # Write netcdf data var contents into array.
    with tiledb.open(array_filename, 'w', ctx=ctx) as A:
        A[write_indices] = data_var[...]


def write_multiattr_array(array_filename, data_vars,
                          start_index=None, scalar=False, ctx=None):
    """Write to each attr in the array."""
    # Determine shape of items to be written.
    zeroth_key = list(data_vars.keys())[0]
    shape = data_vars[zeroth_key].shape  # All data vars *must* have the same shape for writing...
    if scalar:
        shape = (1,) + shape

    # Get write indices.
    if start_index is None:
        start_index = 0
        write_indices = _array_indices(shape, start_index)
    else:
        write_indices = start_index

    # Check for attrs with no data.
    for name, data_var in data_vars.items():
        if data_var is None:
            # Handle missing data for this attr.
            missing_data = np.empty(shape)
            missing_data.fill(np.nan)
            data_vars[name] = missing_data

    # Write netcdf data var contents into array.
    with tiledb.open(array_filename, 'w', ctx=ctx) as A:
        A[write_indices] = {name: data_var[...] for name, data_var in data_vars.items()}


def _dim_inds(dim_points, spatial_inds, offset=0):
    """Convert coordinate values to index space."""
    return [list(dim_points).index(si) + offset for si in spatial_inds]


def _dim_points(points):
    """Convert a dimension variable (coordinate) points to index space."""
    points_ndim = points.ndim
    if points_ndim != 0:
        emsg = f"The append dimension must be 1D, got {points_ndim}D array."
        raise ValueError(emsg)
    start, stop = points[0], points[-1]
    step, = np.unique(np.diff(points))
    return start, stop, step


def _dim_offsets(dim_points, self_stop_ind, self_stop, self_step,
                 scalar=False, points_offset=None):
    """
    Calculate the offset along a dimension given by `var_name` between self
    and other.

    """
    if scalar:
        other_start = dim_points
        spatial_inds = [other_start, other_start]  # Fill the nonexistent `stop` with a blank.
    else:
        other_start, other_stop, other_step = _dim_points(dim_points)
        assert self_step == other_step, "Step between coordinate points is not equal."
        spatial_inds = [other_start, other_stop]

    if points_offset is None:
        points_offset = other_start - self_stop
    inds_offset = int(points_offset / self_step) + self_stop_ind

    i_start, _ = _dim_inds(dim_points, spatial_inds, inds_offset)
    return i_start


def fillnan(xi, y0, diff):
    """A simple linear 1D interpolator."""
    return y0 + (xi * diff)


def _progress_report(other_data_model, verbose, i, total):
    """A helpful printout of append progress."""
    # XXX Not sure about this when called from a bag...
    if verbose and i is not None and total is not None:
        fn = other_data_model.netcdf_filename
        ct = i + 1
        pc = 100 * (ct / total)
        print(f'Processing {fn}...  ({ct}/{total}, {pc:0.1f}%)', end="\r")


def _make_multiattr_tile(other_data_model, domain_path, data_array_name,
                         var_names, append_axis, append_dim, scalar_coord,
                         self_ind_stop, self_dim_stop, self_step,
                         scalar_offset=None, do_logging=False, ctx=None):
    """Process appending a single tile to `self`, per domain."""
    other_data_vars = {}
    for hashed_name in var_names:
        result = other_data_model.data_vars_mapping.get(hashed_name, None)
        if result is not None:
            data_var_name = result[1]
            other_data_vars[hashed_name] = other_data_model.variables[data_var_name]
        else:
            other_data_vars[hashed_name] = None

    # Raise an error if no match in data vars between existing array and other_data_model.
    if len(list(other_data_vars.keys())) == 0:
        emsg = "Variable names in data model [{}] not present in existing array."
        raise KeyError(emsg.format(', '.join(other_data_model.data_var_names)))

    zeroth_data_var = list(other_data_vars.keys())[0]
    data_var_shape  = other_data_vars[zeroth_data_var].shape
    other_dim_var = other_data_model.variables[append_dim]
    other_dim_points = np.atleast_1d(other_dim_var[:])

    # Check for the dataset being scalar on the append dimension.
    if not scalar_coord and len(other_dim_points) == 1:
        scalar_coord = True

    if scalar_coord:
        shape = [1] + list(data_var_shape)
    else:
        shape = data_var_shape

    offsets = []
    offset = _dim_offsets(
        other_dim_points, self_ind_stop, self_dim_stop, self_step,
        scalar=scalar_coord, points_offset=scalar_offset)
    offsets = [0] * len(shape)
    offsets[append_axis] = offset
    offset_inds = _array_indices(shape, offsets)
    domain_name = domain_path.split('/')[-1]
    if do_logging:
        logging.error(f'Indices for {other_data_model.netcdf_filename!r} ({domain_name}): {offset_inds}')

    # Append the data from other.
    data_array_path = f"{domain_path}{data_array_name}"
    write_multiattr_array(data_array_path, other_data_vars,
                          start_index=offset_inds, ctx=ctx)
    # Append the extra dimension points from other.
    dim_array_path = f"{domain_path}{append_dim}"
    write_array(dim_array_path, other_dim_var,
                start_index=offset_inds[append_axis], ctx=ctx)

    # I think this got added spuriously...
    # dim_array_path = f"{domain_path}{append_dim}"
    # write_array(dim_array_path, other_dim_var,
    #             start_index=offset_inds[append_axis], ctx=ctx)


def _make_multiattr_tile_helper(serialized_job):
    """
    Helper function to collate the processing of each file in a multi-attr append.

    """
    # Deserialize job args.
    job_args = AppendArgs(**json.loads(serialized_job))
    if job_args.ctx is not None:
        ctx = tiledb.Ctx(config=tiledb.Config(job_args.ctx))
    else:
        ctx = None

    do_logging = False
    if job_args.logfile is not None:
        do_logging = True
        logging.basicConfig(filename=job_args.logfile,
                            level=logging.ERROR,
                            format='%(asctime)s %(message)s',
                            datefmt='%d/%m/%Y %H:%M:%S')

    domains_mapping = job_args.mapping
    domain_paths = job_args.domain
    append_dim = job_args.dim
    append_axes = job_args.axis
    group = job_args.group

    # Record what we've processed...
    if do_logging:
        logging.error(f'Processing {job_args.other!r} ({job_args.job_number+1}/{job_args.n_jobs})')

    # To improve fault tolerance all the append processing happens in a try/except...
    try:
        if group:
            # Make an NCDataModelGroup from the list of paths in `other`.
            odms = []
            for fn in other:
                odm = NCDataModel(fn)
                odm.populate()
                odms.append(odm)
            other_data_model = NCDataModelGroup(odms)
        elif isinstance(job_args.other, NCDataModel):
            other_data_model = job_args.other
        else:
            other_data_model = NCDataModel(job_args.other)
            other_data_model.populate()

        with other_data_model.open_netcdf():
            for n, domain_path in enumerate(domain_paths):
                if job_args.verbose:
                    fn = other_data_model.netcdf_filename
                    job_no = job_args.job_number
                    n_jobs = job_args.n_jobs
                    n_domains = len(domain_paths)
                    print(f'Processing {fn}...  ({job_no+1}/{n_jobs}, domain {n+1}/{n_domains})', end="\r")

                append_axis = append_axes[n]
                if domain_path.endswith('/'):
                    _, domain_name = os.path.split(domain_path[:-1])
                else:
                    _, domain_name = os.path.split(domain_path)
                array_var_names = domains_mapping[domain_name]
                _make_multiattr_tile(other_data_model, domain_path, job_args.name,
                                     array_var_names, append_axis, append_dim, job_args.scalar,
                                     job_args.ind_stop, job_args.dim_stop, job_args.step,
                                     scalar_offset=job_args.offset, do_logging=do_logging, ctx=ctx)
    except Exception as e:
        emsg = f'Could not process {job_args.other!r}. Details:\n{e}\n'
        logging.error(emsg, exc_info=True)
        if job_args.logfile is None and job_args.verbose:
            raise

def _make_tile(other, domain_path, var_name, append_axis, append_dim,
               self_ind_stop, self_dim_stop, self_step,
               make_data_model, verbose, i=None, num=None):
    """Process appending a single tile to `self`."""
    if make_data_model:
        other_data_model = NCDataModel(other)
        other_data_model.classify_variables()
        other_data_model.get_metadata()
    else:
        other_data_model = other

    _progress_report(other_data_model, verbose, i, num)

    other_data_var = other_data_model.variables[var_name]
    other_dim_var = other_data_model.variables[append_dim]
    other_dim_points = np.atleast_1d(other_dim_var[:])

    # Check for the dataset being scalar on the append dimension.
    scalar_coord = False
    if len(other_dim_points) == 1:
        scalar_coord = True

    if scalar_coord:
        shape = [1] + list(other_data_var.shape)
    else:
        shape = other_data_var.shape

    offsets = []
    try:
        offset = _dim_offsets(
            other_dim_points, self_ind_stop, self_dim_stop, self_step,
            scalar=scalar_coord)
        offsets = [0] * len(shape)
        offsets[append_axis] = offset
        offset_inds = _array_indices(shape, offsets)
    except Exception as e:
        logging.info(f'{other_data_model.netcdf_filename} - {e}')

    # Append the data from other.
    data_array_path = os.path.join(domain_path, var_name)
    write_array(data_array_path, other_data_var, start_index=offset_inds)
    # Append the extra dimension points from other.
    dim_array_path = os.path.join(domain_path, append_dim)
    write_array(dim_array_path, other_dim_var, start_index=offset_inds[append_axis])


def _make_tile_helper(args):
    """Helper method to call from a `map` operation and unpack the args."""
    _make_tile(*args)



import os


class ArrayPath(object):
    def construct_path(self, group_name, array_name):
        raise NotImplementedError


class PosixArrayPath(ArrayPath):
    def __init__(self, base_path, tiledb_name):
        self.base_path = base_path
        self.tiledb_name = tiledb_name

        self.basename = os.path.join(self.base_path, self.tiledb_name)

    def construct_path(self, group_name, array_name):
        return os.path.join(self.basename, group_name, array_name)


class AzureArrayPath(ArrayPath):
    def __init__(self, container, tiledb_name, ctx):
        self.container = container
        self.tiledb_name = tiledb_name
        self.ctx = ctx

        self.basename = f'azure://{self.container}/{self.tiledb_name}'

    def construct_path(self, group_name, array_name):
        return f'{self.basename}/{group_name}/{array_name}'


class AWSArrayPath(ArrayPath):
    def __init__(self):
        raise NotImplementedError


import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tiledb_netcdf",
    version="0.2.3",
    author="Peter Killick",
    author_email="peter.killick@informaticslab.co.uk",
    description="An adapter to convert NetCDF files to TileDB or Zarr arrays",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/informatics-lab/tiledb_netcdf",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "distributed>=2.28.0",
        "tiledb>=0.6.6",
        "scitools-iris>=2.4.0",
        "xarray>=0.15.1",
        "zarr>=2.4.0"
    ]
)
